//=============================
    
    function init_pzem_dialogues(){
        
        $( "#addpzem_dialog" ).dialog({
            autoOpen:   false,
            width:      400,
            height:     450,
            close: function(){
                $("#addpzem_name").val('');
                $("#addpzem_descr").val('');
                $("#addpzem_ports").empty();
                $("#addpzem_devtypes").empty();
                $("#addpzem_logdb").prop('checked',false);
                $("#addpzem_clrdb").val('0');

            },
            open:   function(){
                $("#addpzem_name").val('');
                $("#addpzem_descr").val('');
                $("#addpzem_ports").empty();
                $("#addpzem_devtypes").empty();
                $("#addpzem_resp_msg").html("");
                $("#addpzem_logdb").prop('checked',false);
                $("#addpzem_clrdb").val('0');
                $("#addpzem_clrdb").prop('disabled', true);
                
                db.portlist.toArray()
                    .then(objects => {
                        $("#addpzem_ports").append('<option value="0">Select a port</option>');
                        $.each( objects, function( index, portobj ) {
                            $("#addpzem_ports").append('<option value="' + portobj.id + '">['+ portobj.id+'] ' + portobj.port_name + '</option>');
                        });

                        $("#addpzem_ports").val("0").change();
                        
                        $("#addpzem_devtypes").append('<option value="0">Select device type</option>');
                        $("#addpzem_devtypes").append('<option value="pzem-004t v2.0">pzem-004t v2.0</option>');
                        $.each( Object.keys(devTypesRS485), function( index, devtype ) {
                            $("#addpzem_devtypes").append('<option value="' + devtype + '">' + devtype + '</option>');
                        });
                        _save_changes();
                        
                    },err => {
                        show_err( "Error while reading ports from DB: " + err );
                    });
                

                function _save_changes(){ // call this only when #addpzem_ports <select> is ready
                
                    var selrow = $("#addpzem_dialog").data('selrow');
                    var mode = $("#addpzem_dialog").data('mode');
                    
                    
                    if ( mode == "edit" ){
                        
                        db.devlist.get(selrow )
                            .then( db_record => {
                                $("#addpzem_name").val( db_record.pzem_name );
                                $("#addpzem_descr").val( db_record.pzem_descr );
                                $("#addpzem_logdb").prop('checked',db_record.logDB?true:false);
                                $("#addpzem_clrdb").val(db_record.delDB);
                                
                                if( $("#addpzem_logdb").prop('checked') ){ 
                                    $("#addpzem_clrdb").prop('disabled', false);
                                } else{
                                    $("#addpzem_clrdb").prop('disabled', true);
                                }
                                
                                
                                if ( $("#addpzem_ports").find('option[value="' + db_record.pzem_port + '"]').length === 0){
                                    show_err( "Warning! Port Id " + db_record.pzem_port + " which is assigned to the device Id " + selrow + " - does not exist!");
                                }else{
                                    $("#addpzem_ports").val( db_record.pzem_port ).change();
                                }
                                
                                $("#addpzem_devtypes").val( db_record.devtype ).change();
                                
                            },err => {
                                $("#addpzem_resp_msg").html("<div style='color:#a93226'>Error while editing device: " + err + " </div>");
                            });
    
                        $('#addpzem_addbutton_id').button('option', 'label', 'Save');
                    }else{
                        $('#addpzem_addbutton_id').button('option', 'label', 'Add');;
                    }
                }
                
                
            },
            buttons: [
                {
                    id:     'addpzem_addbutton_id',
                    text:   "",
                    click: function() {
                        $("#addpzem_resp_msg").html("");
                        
                        pzname  = $("#addpzem_name").val();
                        pzdescr = $("#addpzem_descr").val();
                        pzport  = parseInt( $("#addpzem_ports").val() );
                        logDB   = $('#addpzem_logdb').prop('checked');              
                        delDB   = parseInt( $("#addpzem_clrdb").val() );
                        devtype = $("#addpzem_devtypes").val()
                        
                        if( !Number.isInteger(pzport) ){
                            show_err( "Please select a serial port");
                            return;
                        }
                        
                        if( devtype == ""){
                            show_err( "Please select a device type");
                            return;
                        }
                        
                        if( devtype != "pzem-004t v2.0" && devTypesRS485[devtype] === undefined ){
                            show_err( "Unknown device type");
                            return;
                        }
                        
                        
                        
                        if(!Number.isInteger(delDB) || delDB < 1 ) delDB = 0;

                        var selrow = $("#addpzem_dialog").data('selrow');
                        var mode = $("#addpzem_dialog").data('mode');
                  
                        var db_record = {};
                        var pz_settings =   {
                                                readVoltage:        false, 
                                                readCurrent:        false, 
                                                readPower:          false, 
                                                readEnergy:         false, 
                                                readFrequency:      false, 
                                                readPowerfactor:    false,
                                                pollDelay:          5,
                                                nrjAdjust:          0,
                                                slaveAddr:          248
                                            };
                                               
                        
                        (async function() {

                            var port = await db.portlist.get(pzport).catch((err) => { show_err( "Error while reading ports from DB: " + err ); });

                            if( port != undefined && port.port_type == "RS232"){

                                var devices = await db.devlist.toArray().catch((err) => { show_err( "Error while reading devices from DB: " + err ); });
                                var dev = devices.find((obj) => { return obj["pzem_port"] == pzport });
                                
                                if( dev != undefined && dev.id != selrow){
                            
                                    $("#addpzem_resp_msg").html("<div style='color:#a93226'>Port Id " + port.id + " is already assigned to the device Id " + dev.id + " RS232 port can be assigned only once.<div>");
                                    return;
                                }
                            }
                        
                            if ( mode == "edit" ){
                                var record = await db.devlist.get(selrow ).catch((err) => { show_err("Error while getting port settings " + err); });
                                pz_settings = record.pzem_settings;
                                db_record = {id: selrow, pzem_name: pzname, pzem_descr: pzdescr, pzem_port: pzport, pzem_settings: pz_settings, logDB: logDB, delDB: delDB, devtype: devtype};
                            }else{
                                db_record = {pzem_name: pzname, pzem_descr: pzdescr, pzem_port: pzport, pzem_settings: pz_settings, logDB: logDB, delDB: delDB, devtype: devtype};
                            }
                            
                            
                            await db.devlist.put(db_record)
                                .then( id => {
                                    $("#addpzem_resp_msg").html("<div style='color:#117a65'>Successfully " + ((mode=="edit")?"edited":"added") + " device ID " + id + "</div>");
                                    $("#pzemtable").trigger("reloadGrid");
                                },err => {
                                    $("#addpzem_resp_msg").html("<div style='color:#a93226'>Error while " + ((mode=="edit")?"editing":"adding") +  "a new device: " + err + " </div>");
                                });
                        })();
  
                    }
                },
                {
                    text: "Close",
                    click: function() {
                        $("#addpzem_resp_msg").html(""),
                        $( this ).dialog( "close" );
                    }
                }
                
            ]
        });
        
        $( "#delpzem_dialog" ).dialog({
            autoOpen:   false,
            width:      400,
            height:     200,
            close: function(){
                $("#delpzem_consent_msg").val('');
                $("#delpzem_resp_msg").val('');
            },
            open: function(){
                $("#delpzem_resp_msg").val('');
                $("#delpzem_consent_msg").html("Do you really want to <span style='color:red;'> delete </span>PZEM device <span style='color:red'>Id " + $("#delpzem_dialog").data('selrow') + "</span> and all its data from database?");
            },
            buttons: [
                {
                    text: "Delete",
                    click: function() {
                        
                        var deviceid = $("#delpzem_dialog").data('selrow');
                        
                        db.devlist.delete(deviceid)
                            .then( id => { // id always undefined
                                console.log("deleting records for id: " + deviceid);
                                db.records
                                    .where("devid").equals( deviceid )
                                    .delete()
                                    .then( deleteCount => {
                                        console.log( "Deleted db records" + deleteCount);
                                    }, err => {
                                        console.log( "Delete error: " + err);
                                    });
    
                                $("#delpzem_consent_msg").html("");
                                $("#pzemtable").trigger("reloadGrid");
                                $( this ).dialog( "close" );
                            },err => {
                                 $("#delpzem_resp_msg").html("<div style='color:#a93226'>Error while deleting device: " + err + " </div>");
                            });
                    }
                },
                {
                    text: "Cancel",
                    click: function() {
                        $("#delpzem_consent_msg").html("");
                        $( this ).dialog( "close" );
                    }
                }                
            ]
 
        });
        
        
        $( "#configure_RS232_dialog" ).dialog({
            autoOpen:   false,
            width:      400,
            height:     350,
            close: function(){
                $("#configure_RS232_voltage").prop('checked',false);
                $("#configure_RS232_current").prop('checked',false);
                $("#configure_RS232_power").prop('checked',false);
                $("#configure_RS232_energy").prop('checked',false);
                
                $("#configure_RS232_polldelay").val(5);
                $("#configure_RS232_nrjadj").val(0);
                
            },
            open: function(){
                
                var settings = $( "#configure_RS232_dialog" ).data('settings');
                var device_id = $( "#configure_RS232_dialog" ).data('selrow');
                
                
                $("#configure_RS232_voltage").prop('checked',settings.readVoltage?true:false);
                $("#configure_RS232_current").prop('checked',settings.readCurrent?true:false);
                $("#configure_RS232_power").prop('checked',settings.readPower?true:false);
                $("#configure_RS232_energy").prop('checked',settings.readEnergy?true:false);
                
                if(!Number.isInteger(settings.pollDelay) || settings.pollDelay < 1 ) settings.pollDelay = 6;
                if(!Number.isInteger(settings.nrjAdjust) || settings.nrjAdjust < 0 ) settings.nrjAdjust = 0;
                
                $("#configure_RS232_polldelay").val(settings.pollDelay);
                $("#configure_RS232_nrjadj").val(settings.nrjAdjust);
                
                $( "#configure_RS232_dialog" ).dialog({ title: "Configure connection settings for device Id " + device_id });
                $("#configure_RS232_resp_msg").html("");
            },
            buttons: [
                {
                    text: "Save",
                    click: function() {
                        
                        var device_id = parseInt( $( "#configure_RS232_dialog" ).data('selrow') );
                        
                        
                        var pzem_polldelay  =  parseInt( $("#configure_RS232_polldelay").val() );
                        if(isNaN(pzem_polldelay) || pzem_polldelay < 1 ) {
                            $("#configure_RS232_resp_msg").html("<div style='color:#a93226'>Incorrect value for Polling delay</div>");
                            return;
                        }
                        
                        if( pzem_polldelay % 50 ){
                            $("#configure_RS232_resp_msg").html("<div style='color:#a93226'>Polling delay must be a multiple of 50ms</div>");
                            return;
                        }
                        
                        var pzem_nrjadj     =  parseInt( $("#configure_RS232_nrjadj").val() );
                        if(isNaN(pzem_nrjadj) || pzem_nrjadj < 0 ) {
                            $("#configure_RS232_resp_msg").html("<div style='color:#a93226'>Incorrect value for Energy adjust</div>");
                            return;
                        }
                        
                        /*
                        var count = $("#configure_RS232_voltage, #configure_RS232_current, #configure_RS232_power, #configure_RS232_energy").filter(':checked').length;
                        
                        var min_delay = Math.ceil(count * pzem_wrdelay/1000);
            
                        if ( pzem_polldelay < min_delay ){
                            show_err("min polling delay = Write delay * number of selected parameters (" + min_delay + " sec)");
                            return;
                        } 
                         */
                        db.devlist.get( device_id )
                                .then( db_record => {
                                    

                                    db_record.pzem_settings.readVoltage = $('#configure_RS232_voltage').prop('checked');
                                    db_record.pzem_settings.readCurrent = $('#configure_RS232_current').prop('checked');
                                    db_record.pzem_settings.readPower   = $('#configure_RS232_power').prop('checked');
                                    db_record.pzem_settings.readEnergy  = $('#configure_RS232_energy').prop('checked');
                                    db_record.pzem_settings.pollDelay   = pzem_polldelay;
                                    db_record.pzem_settings.nrjAdjust   = pzem_nrjadj;
                                    
                                    db.devlist.put( db_record )
                                        .then( id => {
                                            $("#configure_RS232_resp_msg").html("<div style='color:#117a65'>Settings for device ID " + id + " are saved.</div>");
                                        },err => {
                                            $("#configure_RS232_resp_msg").html("<div style='color:#a93226'>DB Error while updating device settings</div>");
                                        });
                            
                                },err => {
                                    $("#configure_RS232_resp_msg").html("<div style='color:#a93226'>Error while getting device from DB for update</div>");
                                    return;
                                });
                     }
                },
                {
                    text: "Close",
                    click: function() {
                        $("#addpzem_resp_msg").html(""),
                        $( this ).dialog( "close" );
                    }
                }
                
            ]
        });
        
        $( "#configure_RS485_dialog" ).dialog({
            autoOpen:   false,
            width:      512,
            height:     400,
            close: function(){
                $("#configure_RS485_voltage").prop('checked',false);
                $("#configure_RS485_current").prop('checked',false);
                $("#configure_RS485_power").prop('checked',false);
                $("#configure_RS485_energy").prop('checked',false);
                $("#configure_RS485_frequency").prop('checked',false);
                $("#configure_RS485_powerfactor").prop('checked',false);
                
                $("#configure_RS485_polldelay").val(5);
                $("#configure_RS232_currslaveaddr").val(248);
                $("#configure_RS232_newslaveaddr").val(1);
                $("#configure_RS485_chaddr, #configure_RS485_energyreset").unbind();
                
            },
            open: function(){
                var device_id   = parseInt( $( "#configure_RS485_dialog" ).data('selrow') );
                var device      = $( "#configure_RS485_dialog" ).data('device');
                var settings    = device.pzem_settings;
                var port_id     = device.pzem_port;
                
                $("#configure_RS485_voltage").prop('checked',settings.readVoltage?true:false);
                $("#configure_RS485_current").prop('checked',settings.readCurrent?true:false);
                $("#configure_RS485_power").prop('checked',settings.readPower?true:false);
                $("#configure_RS485_energy").prop('checked',settings.readEnergy?true:false);
                $("#configure_RS485_frequency").prop('checked',settings.readFrequency?true:false);
                $("#configure_RS485_powerfactor").prop('checked',settings.readPowerfactor?true:false);

                if(!Number.isInteger(settings.pollDelay) || settings.pollDelay < 1 ) settings.pollDelay = 500;
                if(!Number.isInteger(settings.slaveAddr) || settings.slaveAddr < 1 || settings.slaveAddr > 248) settings.slaveAddr = 248;
                
                $("#configure_RS485_polldelay").val(settings.pollDelay);
                $("#configure_RS485_currslaveaddr").val(settings.slaveAddr);
                 
                $( "#configure_RS485_dialog" ).dialog({ title: "Configure connection settings for device Id " + device_id });
                $("#configure_RS485_resp_msg").html("");
                
                var cur_port = serialPorts[port_id];
                
                
                $("#configure_RS485_chaddr, #configure_RS485_energyreset").click(function(){
                
                    if( cur_port['isOpened'] == false ){
                        $("#configure_RS485_resp_msg").html("<div style='color:#a93226'>Device's port is closed<div>");
                        return;
                    }
                    
                    var cmd = ((this.id =='configure_RS485_chaddr') ? 1 : (this.id =='configure_RS485_energyreset') ? 2 : 0);

                    if( cmd == 0 ){
                        $("#configure_RS485_resp_msg").html("<div style='color:#a93226'> Unknown event source <div>");   
                        return;
                    }
                    
                    
                    
                    $("#configure_RS485_resp_msg").html("<div style='color:#117a65'>Issuing command to " + ((cmd == 1)?"change slave address":"reset energy counter") + "...</div>");
                    var newaddr = parseInt( $("#configure_RS485_newslaveaddr").val() );
                    delayms( 500 );
                    
                    if(cmd == 1){
                        cur_port.writeQueue.push([device_id, settings.slaveAddr, 0xA2, newaddr]);  // set new slave address
                    }else{
                        cur_port.writeQueue.push([device_id, settings.slaveAddr, 0xA3 ]);          // reset energy
                    }
                    
                    // check - if port of our device is in queue
                    /*
                    var port_in_queue = Object.keys(PZEMdevices).find(devid => ((PZEMdevices[devid]['device']['pzem_port'] == cur_port.portid) && (PZEMdevices[devid]['active'] == true)));
                   
                    if( port_in_queue == undefined ){ // port is not in queue so we must trigger relevant subroutines from here
                        if( !cur_port.isBusy ){ 
                            get_RS485_packet( device_id, device )
                                .catch( err=>{ 
                                        console.error(err); 
                                        $("#configure_RS485_resp_msg").html("<div style='color:#a93226'>" + ((cmd == 1)?"Address changing":"Energy reset") +" failed: " + err + "<div>");
                                        cur_port.isBusy = false; 
                                        cur_port.writer.releaseLock(); 
                                    });
                        }else{
                            $("#configure_RS485_resp_msg").html("<div style='color:#a93226'>Operation failed: port is busy.</div>");
                        }
                    }else{
                        // what else?
                    }
                    */
                });
                
            },
            buttons: [
                {
                    text: "Save",
                    click: function() {
                        
                        var device_id = parseInt( $( "#configure_RS485_dialog" ).data('selrow') );
                        
                        db.devlist.get( device_id )
                                .then( db_record => {
                                    var pzem_polldelay  =  parseInt( $("#configure_RS485_polldelay").val() );
                                    var pzem_slaveaddr  =  parseInt( $("#configure_RS485_currslaveaddr").val() );
                            
                                    if(isNaN(pzem_polldelay) || pzem_polldelay < 1 ) {
                                        $("#configure_RS232_resp_msg").html("<div style='color:#a93226'>Incorrect value for Polling delay</div>");
                                        return;
                                    }
                                    if( pzem_polldelay % 50 ){
                                        $("#configure_RS485_resp_msg").html("<div style='color:#a93226'>Polling delay must be a multiple of 50ms</div>");
                                        return;
                                    }
                                    if(isNaN(pzem_slaveaddr) || pzem_slaveaddr < 1 || pzem_slaveaddr > 248  ){
                                        $("#configure_RS485_resp_msg").html("<div style='color:#a93226'>Incorrect value for Slave address</div>");
                                        return;  
                                    }
                            
                                    db_record.pzem_settings.readVoltage         = $('#configure_RS485_voltage').prop('checked');
                                    db_record.pzem_settings.readCurrent         = $('#configure_RS485_current').prop('checked');
                                    db_record.pzem_settings.readPower           = $('#configure_RS485_power').prop('checked');
                                    db_record.pzem_settings.readEnergy          = $('#configure_RS485_energy').prop('checked');
                                    db_record.pzem_settings.readFrequency       = $('#configure_RS485_frequency').prop('checked');
                                    db_record.pzem_settings.readPowerfactor     = $('#configure_RS485_powerfactor').prop('checked');
                                    db_record.pzem_settings.pollDelay           = pzem_polldelay;
                                    db_record.pzem_settings.slaveAddr           = pzem_slaveaddr;
                                    
                                    db.devlist.put(db_record)
                                        .then( id => {
                                            $("#configure_RS485_resp_msg").html("<div style='color:#117a65'>Settings for device ID " + id + " are saved.</div>");
                                        },err => {
                                            $("#configure_RS485_resp_msg").html("<div style='color:#a93226'>DB Error while updating device settings</div>");
                                        });
                            
                                },err => {
                                    $("#configure_RS485_resp_msg").html("<div style='color:#a93226'>Error while getting device from DB for update</div>");
                                    return;
                                });
                    }
                },
                {
                    text: "Close",
                    click: function() {
                        $("#addpzem_resp_msg").html(""),
                        $( this ).dialog( "close" );
                    }
                }
                
            ]
        });
        
        // each parameter is requested separately from RS232 PZEM
        // and we need about 500ms delay between sequential requests
        $("#configure_RS232_voltage, #configure_RS232_current, #configure_RS232_power, #configure_RS232_energy").on('change', function() {
            var count = $("#configure_RS232_voltage, #configure_RS232_current, #configure_RS232_power, #configure_RS232_energy").filter(':checked').length;
            var poll_delay = parseInt( $("#configure_RS232_polldelay").val() );
            var min_delay = Math.ceil(count * 0.5);
           
            if( poll_delay < min_delay ) $("#configure_RS232_polldelay").val( min_delay )  ;
            
        });
        
        $("#addpzem_logdb").on('change', function(){
            if( $(this).prop('checked') ){ 
                $("#addpzem_clrdb").prop('disabled', false);
            } else{
                $("#addpzem_clrdb").prop('disabled', true);
            }
        });
        
        // just a visual embellishment. just to show that the only type of RS232 devices we know
        // is pzem-004t v2.0 and this device has a dedicated communication function
        // 
        $("#addpzem_ports").on('change', function(){
            var portid = $(this).val();
            if( portid == 0 ) return;
            var type = serialPorts[portid]['type'];
            if( type == "RS232"){
                $("#addpzem_devtypes").val("pzem-004t v2.0").change();
                $("#addpzem_devtypes").prop('disabled', true);
            }else{
                $("#addpzem_devtypes").prop('disabled', false);
            }
        });
        
        
    }
    