//=============================

    function init_pzem_table(){
    
        $("#pzemtable").jqGrid({ 
            datatype: function(postdata){

                db.table("devlist").toArray()
                .then( dbdevices => {
                    
                    // remove from local object devices which does not exist in the database
                    Object.keys(PZEMdevices).forEach( devid => {
                        if ( dbdevices.find((obj) => { return obj["id"] == devid }) === undefined ){
                            
                            if( PZEMdevices[devid]['active'] ){
                                show_err( "Device is absent is DB, but present in local copy and marked as active Id " + devid );
                                return;
                            }
                            
                            delete PZEMdevices[devid];
                        }
                    });
                    
                    // update port properties from database
                    $.each( dbdevices, function( index, device ) {
                        
                        var old_device = PZEMdevices[device.id];
                        
                        if( old_device === undefined ){ // create a new object
                            
                            PZEMdevices[device.id] = {
                                device:         device,        
                                port_type:      serialPorts[ device.pzem_port ]['type'], // serialPorts hopefully initialized already
                                counter:        0,             
                                pkt_cntr:       0,             
                                err_cntr:       0,             
                                active:         false          
                            }
                            
                        }else{
                            PZEMdevices[device.id]['device'] = device;
                            PZEMdevices[device.id]['port_type'] = serialPorts[ device.pzem_port ]['type'];
                        }
                        
                        device.pzem_connected = PZEMdevices[device.id]['active'] ? 1 : 0;
                        
                    });

                    $("#pzemtable")[0].addJSONData(dbdevices);
   
                }, err =>{
                    show_err( "Error while reading devices from DB: " + err );
                });
            },
            width:            1024,
            shrinkToFit:      true,
            colNames:['Id', 'Name', 'Description', 'Port', 'Voltage', 'Current', 'Power', 'Energy', 'Freq', 'cos &#966;','last update',''], 
            colModel:[ 
                {name:'id',             index:'id',             width:20,   sortable:false,  editable:false,    title:false},
                {name:'pzem_name',      index:'pzem_name',      width:100,  sortable:false,  editable:false,    title:false},
                {name:'pzem_descr',     index:'pzem_descr',     width:190,  sortable:false,  editable:false,    title:false},
                {name:'pzem_port',      index:'pzem_port',      width:100,  sortable:false,  editable:false,    title:false, formatter: portIdNameFmatter},
                {name:'pzem_curv',      index:'pzem_curv',      width:50,   sortable:false,  editable:false,    title:false},
                {name:'pzem_curi',      index:'pzem_curi',      width:50,   sortable:false,  editable:false,    title:false},
                {name:'pzem_curp',      index:'pzem_curp',      width:50,   sortable:false,  editable:false,    title:false},
                {name:'pzem_cure',      index:'pzem_cure',      width:70,   sortable:false,  editable:false,    title:false},
                {name:'pzem_curf',      index:'pzem_curf',      width:40,   sortable:false,  editable:false,    title:false},
                {name:'pzem_curpf',     index:'pzem_curpf',     width:50,   sortable:false,  editable:false,    title:false},
                {name:'pzem_lastupd',   index:'pzem_lastupd',   width:245,  sortable:false,  editable:false,    title:false},
                {name:'pzem_connected', index:'pzem_connected', width:0,    sortable:false,  editable:false,    title:false, hidden:true },

                
            ], 
            pginput:        false,
            pgbuttons:      false,
            viewrecords:    false, 
            rows:           100,
            height:         'auto', 
            caption:        "Available devices",
            sortname:       'id', 
            sortorder:      "desc",
            gridview:       true,
            rowattr: function (rd) {
                 if (rd.pzem_connected == 1) { 
                    return {"class": "JqRowGreen"};
                }
            },
            pager:          "#pzempager",
            editurl:        'clientArray'
        });
        
        $("#pzemtable").jqGrid('navGrid','#pzempager', 
            {
                add:        false,
                edit:       false,
                del:        false,
                search:     false,
                refresh:    true  
            },
            {},
            {}, // 'add' options
            {}, /*del options*/
            {}
        );
        
        
        $("#pzemtable").jqGrid('navGrid',"#pzempager").jqGrid('navButtonAdd',"#pzempager",{
            caption:"", 
            buttonicon:"ui-icon-trash", 
            position: "first", 
            title:"Delete PZEM device", 
            cursor: "pointer",
            onClickButton: function(){
            
                var selrow = parseInt( $("#pzemtable").jqGrid ('getGridParam', 'selrow') );
                if( !Number.isInteger( selrow ) ){
                    show_err( 'error: No row selected');
                    return;
                }
                else{
                    $("#delpzem_dialog").data('selrow', selrow).dialog('open');
                }
            }
        });
        
        $("#pzemtable").jqGrid('navGrid',"#pzempager").jqGrid('navButtonAdd',"#pzempager",{
            caption:"", 
            buttonicon:"ui-icon-transferthick-e-w", 
            position: "first", 
            title:"Start/Stop polling", 
            cursor: "pointer",
            onClickButton: function(){
            
                var selrow = parseInt( $("#pzemtable").jqGrid ('getGridParam', 'selrow') );
                if( !Number.isInteger( selrow ) ){
                    show_err( 'error: No row selected');
                    return;
                }
                else{
                        if( PZEMdevices[selrow]['active'] == false ){ // enabling polling
                            
                            if( serialPorts[ PZEMdevices[selrow]['device']['pzem_port'] ] === undefined ){
                                show_err( "Serial port Id " + PZEMdevices[selrow]['device']['pzem_port'] + " doesn't exist.");
                                return;
                            }
                            
                            if( serialPorts[ PZEMdevices[selrow]['device']['pzem_port'] ]['isOpened'] == false ){
                                show_err( "Serial port Id " + PZEMdevices[selrow]['device']['pzem_port'] + " isn't connected.");
                                return;
                            }
                            
                            PZEMdevices[selrow]['counter']  = 0;
                            PZEMdevices[selrow]['pkt_cntr'] = 0;
                            PZEMdevices[selrow]['err_cntr'] = 0;
                            PZEMdevices[selrow]['active']   = true;

                        }else{ // disabling polling
                            
                            PZEMdevices[selrow]['active']   = false;
                            placeCellData(selrow, 'YY', 0); // clear all fields on the table

                        }
                        $("#pzemtable").trigger("reloadGrid");
                    // add device to queue, reload grid to mark device as connected
                }
            }
        });
        
        
        $("#pzemtable").jqGrid('navGrid',"#pzempager").jqGrid('navButtonAdd',"#pzempager",{
            caption:"", 
            buttonicon:"ui-icon-gear", 
            position: "first", 
            title:"Connection settings", 
            cursor: "pointer",
            onClickButton: function(){
                var selrow = parseInt( $("#pzemtable").jqGrid ('getGridParam', 'selrow') );
                if( !Number.isInteger( selrow ) ){
                    show_err( 'error: No row selected');
                    return;
                }
                else{
                    
                    (async function() {
                        var device = await db.devlist.get(selrow).catch((err) => { show_err( "Error while reading device Id " + selrow + " from DB: " + err ); });
                        if( device === undefined ){
                            show_err( "Error while reading device Id "+ selrow +" from DB");
                            return;
                        }
                        
                        var portid = parseInt(device.pzem_port);
        
                        var port = await db.portlist.get(portid).catch((err) => { show_err( "Error while reading port Id " + portid + " from DB: " + err ); });
                        if( port === undefined ){
                            show_err( "Error while reading port Id " + portid + " from DB");
                            return;
                        }
                        
                        if( port.port_type == "RS232" ){
                            $( "#configure_RS232_dialog" ).data( {selrow: selrow, settings: device.pzem_settings} ).dialog('open');
                        }else if( port.port_type == "RS485" ){
                            $( "#configure_RS485_dialog" ).data( {selrow: selrow, device: device} ).dialog('open');
                        }else{
                            show_err( "Unknown port type: " + port.port_type);
                        }
                    })();
                }
            }
        });
        
        $("#pzemtable").jqGrid('navGrid',"#pzempager").jqGrid('navButtonAdd',"#pzempager",{
            caption:"", 
            buttonicon:"ui-icon-pencil", 
            position: "first", 
            title:"Edit PZEM device", 
            cursor: "pointer",
            onClickButton: function(){
                var selrow = parseInt( $("#pzemtable").jqGrid ('getGridParam', 'selrow') );
                if( !Number.isInteger( selrow ) ){
                    show_err( 'error: No row selected' );
                }
                else{
                    $( "#addpzem_dialog" ).dialog({ title: "Edit PZEM device Id " + selrow });
                    $("#addpzem_dialog").data( {'selrow' : selrow, 'mode' : 'edit'} ).dialog('open');
                }
            }
        });
        
        $("#pzemtable").jqGrid('navGrid',"#pzempager").jqGrid('navButtonAdd',"#pzempager",{
            caption:"", 
            buttonicon:"ui-icon-plus", 
            position: "first", 
            title:"Add new PZEM device", 
            cursor: "pointer",
            onClickButton: function(){
                $( "#addpzem_dialog" ).dialog({ title: "Add a new PZEM device"});
                $("#addpzem_dialog").data({'mode': 'add'}).dialog('open');
                
            }
        });
        
        function portIdNameFmatter ( cellvalue, options, rowObject ){
            var portId = parseInt(cellvalue);
            if( !Number.isInteger( portId ) ){
                return cellvalue;
            }
            else{
                var port = serialPorts[portId];
               
                if( port == undefined ){
                    return('[' + cellvalue +'] <span style="color:#F00">Not found!</span>'); 
                }else{
                    return("[" + portId + "] " + port.name);
                }
            }
        }
    }